/** 
 *   Site testing for SKA candidate.
 *
 *   (c) Copyright 2004 Federico Bareilles <fede@iar.unlp.edu.ar>,
 *   Instituto Argentino de Radio Astronomia (IAR).
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License, or (at your option) any later version.
 *     
 *   The author does NOT admit liability nor provide warranty for any
 *   of this software. This material is provided "AS-IS" in the hope
 *   that it may be useful for others.
 *
 *   TODO:
 *        *) Data save.
 *        *)
 *
 *
 **/

#define _GNU_SOURCE

//#define FITS_COMPAT
#define NEW_FILE_FORMAT

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include <sys/file.h>

#include "observ_table.h"
#include "observation.h"
#include "analyzer.h"
#include "my_stdio.h"
#include "my_math.h"
#include "switch_rf.h"
#include "rotator.h"
#include "toska.h"

#define H_PREFIX "# "


static int mkpath(const char *pathname, mode_t mode)
{
	char *dir, *dirp, *d;
	char *path;
	int path_len = 0;
	int len = strlen(pathname) + 1;
	struct stat m;

	if ( stat(pathname, &m) == 0 ) {
		return 0;
	}

	if ((dir = (char *) malloc(len * sizeof(char))) == NULL) {
		return -1;
	}
	if ((path = (char *) malloc(len * sizeof(char))) == NULL) {
		return -1;
	}

	d = dir;
	memcpy(dir, pathname, len);
	dir[len] = 0;
	d = strtok_r(dir,"/", &dirp);
	while (d != NULL) {
		path_len += sprintf(path+path_len, "%s/", d);
		if ( stat(path, &m) != 0 ) {
			if (mkdir(path, mode) != 0) {
				return -1;
			}
		}
		d = strtok_r(NULL,"/",&dirp);
	}
	free(dir);
	free(path);
	
	return 0;
}

	

static int print_observ_header(int fd, struct obs_setup *s)
{
	struct tm *t;
	
	t = localtime(&s->start);
	fd_printf(fd, H_PREFIX"NAXIS   = %d\n", 2);
	fd_printf(fd, H_PREFIX"NAXIS1  = %d\n", s->naxis1);
	fd_printf(fd, H_PREFIX"NAXIS2  = %d\n", 1);
#ifdef FITS_COMPAT
	fd_printf(fd, H_PREFIX"CTYPE1  = 'FREQ'\n");
	fd_printf(fd, H_PREFIX"CDELT1  = %14f\n", 
		  (s->obs->stop_freq - s->obs->start_freq)/
		  ((double) (AN_MAX_POINTS - 1)) * 1e6);
	fd_printf(fd, H_PREFIX"CRVAL1  = %14f\n", 
		  s->obs->start_freq * 1e6);

	fd_printf(fd, H_PREFIX"CTYPE2  = 'db'\n");
	fd_printf(fd, H_PREFIX"CDELT2  = %14f\n", 0.0);
	fd_printf(fd, H_PREFIX"CRVAL2  = %14f\n", 1.0);
#endif
	fd_printf(fd, H_PREFIX"DATE = %04d-%02d-%02d\n",
		  t->tm_year + 1900, t->tm_mon + 1, t->tm_mday);
	fd_printf(fd, H_PREFIX"TIME = %02d:%02d:%02d\n",
		  t->tm_hour, t->tm_min, t->tm_sec);
	fd_printf(fd, H_PREFIX"DATE-UNIX = %ld\n", s->start);
	fd_printf(fd, H_PREFIX"AZIMUTH = %3d\n", s->azimuth); 

	fd_printf(fd, H_PREFIX"MODE = %d\n", s->obs->mode); 
	fd_printf(fd, H_PREFIX"START-FREQ = %8.5f\n", s->obs->start_freq); 
	fd_printf(fd, H_PREFIX"STOP-FREQ  = %8.5f\n", s->obs->stop_freq); 
	fd_printf(fd, H_PREFIX"RBW = %4d\n", s->obs->RBW); 
	fd_printf(fd, H_PREFIX"SWEEP = %4d\n", s->obs->sweep_obs); 
	fd_printf(fd, H_PREFIX"REPETITION = %2d of %2d\n", 
		  s->current_rep, s->obs->repetitions); 
	fd_printf(fd, H_PREFIX"PARTITION = %2d\n", s->partition); 
	fd_printf(fd, H_PREFIX"ANTENNA-CONF = %c\n", s->obs->switchs); 
	fd_printf(fd, H_PREFIX"POLARIZATION = %c\n", s->polarization); 
	fd_printf(fd, H_PREFIX"END\n");

	return 0;
}


static int save_observ(struct obs_setup *s)
{
	char file[PATH_LEN];
#ifdef NEW_FILE_FORMAT
	char dir[PATH_LEN];
#endif
	struct tm *t;
	int fd, i;

/* 

Old:

   s2004112617401030h00959999011000004418A
   |                                      \__________
   |                                                 \
   s 2004 11 26 17 40 1 030 h 00959999 01100000 44 18A
   | |    |  |  |  |  | |   | |        |        |  | +-> Antenna Conf. 
   | |    |  |  |  |  | |   | |        |        |  +---> Current repetition
   | |    |  |  |  |  | |   | |        |        +------> Repetitions
   | |    |  |  |  |  | |   | |        +---------------> Stop freq [Hz]
   | |    |  |  |  |  | |   | +------------------------> Start freq [Hz]
   | |    |  |  |  |  | |   +--------------------------> Polarization
   | |    |  |  |  |  | +------------------------------> Azimuth
   | |    |  |  |  |  +--------------------------------> Mode
   | |    |  |  |  +-----------------------------------> Minutes
   | |    |  |  +--------------------------------------> Hour
   | |    |  +-----------------------------------------> Day
   | |    +--------------------------------------------> Month
   | +-------------------------------------------------> Year
   +---------------------------------------------------> ID 


New:

   1-01/00959999-01100000/030h/

   Directory:
   1 - 01 / 00959999 - 01100000 / 030 h /
   |   |    |          |          |   |
   |   |    |          |          |   +---> Polarization
   |   |    |          |          +-------> Azimuth
   |   |    |          +------------------> Stop freq [Hz]
   |   |    +-----------------------------> Start freq [Hz]
   |   +----------------------------------> Partition
   +--------------------------------------> Mode

   File:
   s11015016004418

   s 1101501600 44 18
   | |          |  |
   | |          |  +-> Current repetition
   | |          +----> Repetitions
   | +---------------> seconds since 00:00:00, Jan 1, 1970
   +-----------------> ID

*/
	t = localtime(&s->start);

#ifdef NEW_FILE_FORMAT

	sprintf(dir,"%s/%1d-%02d/%08.0f-%08.0f/%03d%c",
		s->data_dir, s->mode,s->partition,
		(s->obs->start_freq * 1000000.0),
		(s->obs->stop_freq * 1000000.0),
		s->azimuth, s->polarization);
	mkpath(dir, S_IRWXU | S_IRGRP | S_IXGRP);
	sprintf(file,"%s/s%010ld%02d%02d", 
		dir, s->start, s->obs->repetitions, s->current_rep);

#else
	sprintf(file,"%s/s%04d%02d%02d%02d%02d%1d%03d%c%08.0f%08.0f%02d%02d%c",
		s->data_dir, t->tm_year + 1900, t->tm_mon + 1, t->tm_mday,
		t->tm_hour, t->tm_min, s->mode, s->azimuth, s->polarization,
		(s->obs->start_freq * 1000000.0),
		(s->obs->stop_freq * 1000000.0),
		s->obs->repetitions, s->current_rep, s->obs->switchs);
#endif

	fd = open(file, O_WRONLY|O_CREAT|O_TRUNC, 
		  S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH);
	if (fd < 0) {
		SYSLOG(LOG_ERR,"save_observ: Error opening file: %s", file);
		return -1;
	}
	/* print header */
	print_observ_header(fd, s);
	/* print data */

	for(i=0; i<s->naxis1; i++)
		fd_printf(fd, "%8.2f\n", *(s->data+i));

	close(fd);

	return 0;
	
}


static int get_observ( int dev, struct obs_setup *s)
{
	int ret;
	int i = 0;

	time(&s->start);
	while (analyzer_get_observ( dev, s) != 0) {
		time(&s->start);
		if ((ret = analyzer_reset(dev)) != 0){
		  if (i < AN_RETRY_TIME_OUT) {
		    sleep(1);
		    i++;
		  }
		  else
		    return ret;
		}
	}
	
	return 0;
}


/**
 * Check and set rotator position
 *
 * Return:
 *        -1: current position not equal to setup, try to seting this.
 *         0: position is equal to setup, do nothing.
 *
 **/
static int chk_set_az_pol(struct obs_setup *s)
{
	int az;
	char pol;
	int ret = 0;

	if(! rotator_flag)
	        return 0;

	rotator_get_azimuth(s->fd_rot, &az);
	rotator_get_polarization(s->fd_rot, &pol);
	
	az = (int) grid_norm((double) az, ROTATOR_GRID);
	if ((s->azimuth != az) || (s->polarization != pol)) {
		SYSLOG(LOG_WARNING,
		       "chk_set_az_pol: current %d, %c; spected %d, %c.",
		       az, pol, s->azimuth, s->polarization);
		ret = -1;
		if (s->azimuth != az) 
			rotator_set_azimuth(s->fd_rot, s->azimuth);
		if (s->polarization != pol)
			rotator_set_polarization(s->fd_rot, s->polarization);
	}

	return ret;
}


int observ_mode(int dev, struct obs_setup *s)
{
	struct modes *obsp = s->obs;
	int i, error;
	int done;

	while (obsp != NULL ) {
		if (obsp->mode == s->mode) {
			if (antenna_switchs_set(obsp->switchs) < 0) {
				SYSLOG(LOG_ERR, "Cant set rf switch.");
			}
			for(i = 0; i < obsp->repetitions; i++){
				s->current_rep = i+1;
				s->obs = obsp;
				chk_set_az_pol(s);
				done = 0;
				do {
					error = get_observ( dev, s);
					if (error) {
						return error;
					}
					if (chk_set_az_pol(s) == 0) {
						done = 1;
					}else{
						SYSLOG(LOG_ERR, "rotator position error.");
					}
				} while(!done);
				save_observ(s);
				free(s->data);

			}
		}
		obsp = obsp->next;
	}

	return 0;
}


struct sequence *observ_sequences(int an_dev, 
				 struct obs_setup *s, struct sequence *seq, 
				 struct modes *obs)
{
	int i;
	struct sequence *seqp = seq;
	struct sequence *seqr = seq;
	int level = seqp->level;
	
	while (seqp != NULL && seqp->type != SEQ_TYPE_NULL) {
		if( seqp->level != level) {
			if( seqp->level > level) {
				PVERB(PINFO, 
				      "# loop start %d\n", seqp->level);
				for(i=0; i<seqp->loop; i++) {
					seqr = observ_sequences(an_dev, s, 
								seqp, obs);
				}
				seqp = seqr;
				PVERB(PINFO, "# loop reasume %d\n",level);
				
			}else if( seqp->level < level ){
				PVERB(PINFO, "# loop end %d\n", level);
				return seqp->next;
			}
		}else{
			if ( seqp->type == SEQ_TYPE_DATA) {
				s->obs = obs;
				s->mode = seqp->mode;
				s->azimuth = seqp->azimuth;
				s->polarization = seqp->polarization;
				s->partition = seqp->partition;

				PVERB(PINFO, 
				      "# seq = %d-%02d %3d, %d, %d\n", 
				      seqp->mode, seqp->partition, 
				      seqp->azimuth,
				      seqp->level, seqp->loop);
				
				observ_mode(an_dev, s);
			}		
			if(seqp->type == SEQ_TYPE_RESTART && 
			   level == 0 ) seqp = seq;
			else seqp = seqp->next;
		}
		
	}
	
	return NULL;
}
